from typing import Iterable, Tuple

class Packer:
    def image(self, name: str, image: str | bytes, size: Tuple[int, int] = ...) -> None: ...
    def font(self, name: str, font: str | bytes, size: int, glyphs: Iterable[int] | None = None) -> None: ...
    def build(self, size: Tuple[int, int], oversampling: int = 1, padding: int = 1) -> None: ...

def packer() -> Packer: ...
